/*
   +----------------------------------------------------------------------+
   | PHP Version 4                                                        |
   +----------------------------------------------------------------------+
   | Copyright (c) 1997-2002 The PHP Group                                |
   +----------------------------------------------------------------------+
   | This source file is subject to version 2.02 of the PHP license,      |
   | that is bundled with this package in the file LICENSE, and is        |
   | available at through the world-wide-web at                           |
   | http://www.php.net/license/2_02.txt.                                 |
   | If you did not receive a copy of the PHP license and are unable to   |
   | obtain it through the world-wide-web, please send a note to          |
   | license@php.net so we can mail you a copy immediately.               |
   +----------------------------------------------------------------------+
   | Author: Nikos Mavroyanopoulos <nmav@gnutls.org>                      |
   +----------------------------------------------------------------------+
*/

/* $Id: hydra.c,v 1.8 2002/10/02 12:43:58 nmav Exp $ */

#include "php.h"
#include "SAPI.h"
#include "php_main.h"
#include "php_variables.h"
#include "php_ini.h"

#include "ext/standard/php_smart_str.h"

#include <sys/time.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <stdlib.h>
#include <unistd.h>

#include <hic.h> /* for hic_stuff structure */

typedef struct {
	hic_stuff *hc;
	int read_post_data;	
	void (*on_close)(int);
	long async_send;
} php_hydra_globals;


#ifdef ZTS
static int hydra_globals_id;
#define TG(v) TSRMG(hydra_globals_id, php_hydra_globals *, v)
#else
static php_hydra_globals hydra_globals;
#define TG(v) (hydra_globals.v)
#endif

PHP_INI_BEGIN()
	STD_PHP_INI_ENTRY("async_send", "0", PHP_INI_ALL, OnUpdateInt, async_send, php_hydra_globals, hydra_globals)
PHP_INI_END()

static int sapi_hydra_ub_write(const char *str, uint str_length TSRMLS_DC)
{
	int n;
	uint sent = 0;	
	
	while (str_length > 0) {
		n = write(TG(hc)->out_fd, str, str_length);

		if (n == -1 && errno == EPIPE)
			php_handle_aborted_connection();
		if (n == -1 && errno == EINTR) 
			continue;

		if (n <= 0) 
			return n;

		TG(hc)->bytes_sent += n;
		str += n;
		sent += n;
		str_length -= n;
	}

	return sent;
}

#define COMBINE_HEADERS 30

static int sapi_hydra_send_headers(sapi_headers_struct *sapi_headers TSRMLS_DC)
{
	char buf[1024];
	struct iovec vec[COMBINE_HEADERS];
	int n = 0;
	zend_llist_position pos;
	sapi_header_struct *h;
	size_t len;
	
	if (!SG(sapi_headers).http_status_line) {
		snprintf(buf, sizeof(buf)-1, "HTTP/1.1 %d Something\r\n", SG(sapi_headers).http_response_code);
		len = strlen(buf);
		vec[n].iov_base = buf;
		vec[n].iov_len = len;

	} else {
		vec[n].iov_base = SG(sapi_headers).http_status_line;
		len = strlen(vec[n].iov_base);
		vec[n].iov_len = len;
		vec[++n].iov_base = "\r\n";
		vec[n].iov_len = 2;
		len += 2;
	}
	TG(hc)->status = SG(sapi_headers).http_response_code;
	TG(hc)->bytes_sent += len;
	n++;

        /* Send server version and name
       	 */
	vec[n].iov_base = TG(hc)->server_version;
	vec[n].iov_len = strlen(TG(hc)->server_version);

	TG(hc)->bytes_sent += len;
	n++;

#define DEF_CONTENT_TYPE_LINE "Content-Type: text/html\r\n"
	if (SG(sapi_headers).send_default_content_type) {
		vec[n].iov_base = DEF_CONTENT_TYPE_LINE;
		vec[n].iov_len = sizeof(DEF_CONTENT_TYPE_LINE) - 1;
		n++;
	}

	h = zend_llist_get_first_ex(&sapi_headers->headers, &pos);
	while (h) {
		vec[n].iov_base = h->header;
		vec[n++].iov_len = h->header_len;
		if (n >= COMBINE_HEADERS - 1) {
			/* XXX: partial writevs are not handled */
			if (writev(TG(hc)->out_fd, vec, n) == -1 && errno == EPIPE)
				php_handle_aborted_connection();
			n = 0;
		}
		vec[n].iov_base = "\r\n";
		vec[n++].iov_len = 2;
		
		h = zend_llist_get_next_ex(&sapi_headers->headers, &pos);
	}

	vec[n].iov_base = "\r\n";
	vec[n++].iov_len = 2;

	if (n) {
		/* XXX: partial writevs are not handled */
		if (writev(TG(hc)->out_fd, vec, n) == -1 && errno == EPIPE)
			php_handle_aborted_connection();
	}
	
	return SAPI_HEADER_SENT_SUCCESSFULLY;
}

static int sapi_hydra_read_post(char *buffer, uint count_bytes TSRMLS_DC)
{
	size_t read_bytes = 0, tmp;
	int c;
	int n;

	/* In hydra, post data are written to a file, and an fd of the file is sent here 
	 */

	while (read_bytes < count_bytes) {
		tmp = read(TG(hc)->post_data_fd, buffer + read_bytes, 
				count_bytes - read_bytes);

		if (tmp == 0 || (tmp == -1 && errno != EINTR))
			break;
		/* A simple "tmp > 0" produced broken code on Solaris/GCC */
		if (tmp != 0 && tmp != -1)
			read_bytes += tmp;

	}

	TG(read_post_data) += read_bytes;

	return read_bytes;
}

#define COOKIE "HTTP_COOKIE="
#define COOKIE_SIZE sizeof("HTTP_COOKIE=")-1
static char *sapi_hydra_read_cookies(TSRMLS_D)
{
int i;
char *p = NULL;
        /* register_variables should have been called before this.
         */
        i = 0;
	while( TG(hc)->cgi_env[i] != NULL) {
		if (strncmp( TG(hc)->cgi_env[i], COOKIE, COOKIE_SIZE)==0) {
			p = strchr( TG(hc)->cgi_env[i], '=');
			if (p) p++;
			break;
		}
		i++;
	}
	
	return p;
}

#define BUF_SIZE 512

static void sapi_hydra_register_variables(zval *track_vars_array TSRMLS_DC)
{
	char buf[BUF_SIZE];
	char *p;
	int i;

	php_register_variable("PHP_SELF", SG(request_info).request_uri, track_vars_array TSRMLS_CC);

	buf[BUF_SIZE] = 0;

        /* Register all variables in cgi_env 
         * cgi_env has variables in the form XXX=YYY
         */
        i = 0;
	while( TG(hc)->cgi_env[i] != NULL) {
        	if (TG(hc)->cgi_env[i] == NULL) continue;
		
		strncpy( buf, TG(hc)->cgi_env[i], BUF_SIZE - 1);
		p = strchr( buf, '=');
		if (p!=NULL) { /* replace '=' with '\0' */
			*p=0;
        		php_register_variable( buf, p+1, track_vars_array TSRMLS_CC);
        	}
        	i++;
        }

}

static PHP_MINIT_FUNCTION(hydra)
{
	REGISTER_INI_ENTRIES();
	return SUCCESS;
}

static zend_module_entry php_hydra_module = {
	STANDARD_MODULE_HEADER,
	"hydra",
	NULL,
	PHP_MINIT(hydra),
	NULL,
	NULL,
	NULL,
	NULL, /* info */
	NULL,
	STANDARD_MODULE_PROPERTIES
};

static int php_hydra_startup(sapi_module_struct *sapi_module)
{
	if (php_module_startup(sapi_module) == FAILURE
			|| zend_startup_module(&php_hydra_module) == FAILURE) {
		return FAILURE;
	}
	return SUCCESS;
}

static sapi_module_struct hydra_sapi_module = {
	"hydra",
	"hydra",
	
	php_hydra_startup,
	php_module_shutdown_wrapper,
	
	NULL,									/* activate */
	NULL,									/* deactivate */

	sapi_hydra_ub_write,
	NULL,
	NULL,									/* get uid */
	NULL,									/* getenv */

	php_error,
	
	NULL,
	sapi_hydra_send_headers,
	NULL,
	sapi_hydra_read_post,
	sapi_hydra_read_cookies,

	sapi_hydra_register_variables,
	NULL,									/* Log message */

	NULL,									/* Block interruptions */
	NULL,									/* Unblock interruptions */

	STANDARD_SAPI_MODULE_PROPERTIES
};

static void hydra_module_main(TSRMLS_D)
{
	zend_file_handle file_handle;

	file_handle.type = ZEND_HANDLE_FILENAME;
	file_handle.filename = SG(request_info).path_translated;
	file_handle.free_filename = 0;
	file_handle.opened_path = NULL;

	if (php_request_startup(TSRMLS_C) == FAILURE) {
		return;
	}
	
	php_execute_script(&file_handle TSRMLS_CC);
	php_request_shutdown(NULL);
}

static int hydra_request_ctor(TSRMLS_D)
{
	int offset, i;
	char* query_string;
	char* content_length;
	char* auth = NULL;
	char* request_method;
	char *p = NULL;

	i = 0;
	while( TG(hc)->cgi_env[i] != NULL) {
		if (TG(hc)->cgi_env[i] == NULL) continue;

		if (strncmp( TG(hc)->cgi_env[i], "REQUEST_METHOD=", 15)==0) {
			p = strchr( TG(hc)->cgi_env[i], '=');
			if (p) p++;
			SG(request_info).request_method = p;
		} else 	if (strncmp( TG(hc)->cgi_env[i], "HTTP_AUTHORIZATION=", 19)==0) {
			p = strchr( TG(hc)->cgi_env[i], '=');
			if (p) p++;
			auth = p;
		} else 	if (strncmp( TG(hc)->cgi_env[i], "CONTENT_TYPE=", 13)==0) {
			p = strchr( TG(hc)->cgi_env[i], '=');
			if (p) p++;
			SG(request_info).content_type = p;
		} else 	if (strncmp( TG(hc)->cgi_env[i], "CONTENT_LENGTH=", 15)==0) {
			p = strchr( TG(hc)->cgi_env[i], '=');
			if (p) {
			   p++;
			   SG(request_info).content_length = atoi(p);
			}
		} else 	if (strncmp( TG(hc)->cgi_env[i], "QUERY_STRING=", 13)==0) {
			p = strchr( TG(hc)->cgi_env[i], '=');
			if (p) {
			   p++;
			   SG(request_info).query_string = p;
			}
		} else 	if (strncmp( TG(hc)->cgi_env[i], "REQUEST_URI=", 12)==0) {
			p = strchr( TG(hc)->cgi_env[i], '=');
			if (p) {
			   p++;
			   SG(request_info).request_uri = p;
			}
		} else 	if (strncmp( TG(hc)->cgi_env[i], "PATH_TRANSLATED=", 14)==0) {
			p = strchr( TG(hc)->cgi_env[i], '=');
			if (p) {
			   p++;
			   SG(request_info).path_translated = p;
			}
		}
		i++;
	}
	if ( SG(request_info).path_translated == NULL ||
		SG(request_info).request_uri == NULL) return -1;

	SG(sapi_headers).http_response_code = 200;

	php_handle_auth_data(auth TSRMLS_CC);

	SG(server_context) = (void *) 1;
	
	return 0;
}

static void hydra_request_dtor(TSRMLS_D)
{
	return;
}
                                        


static off_t hydra_php_request(hic_stuff *hc TSRMLS_DC)
{
	TG(hc) = hc;
	hc->bytes_sent = 0;

	TG(read_post_data) = 0;

	if (hydra_request_ctor(TSRMLS_C) == -1)
	   return 0;

	hydra_module_main(TSRMLS_C);

	hydra_request_dtor(TSRMLS_C);

	return 0;
}

off_t _php_hic_request( hic_stuff *hc)
{
	TSRMLS_FETCH();
	return hydra_php_request(hc TSRMLS_CC);
}

void _php_hic_init(void)
{
	char *ini;

#ifdef ZTS
	tsrm_startup(1, 1, 0, NULL);
	ts_allocate_id(&hydra_globals_id, sizeof(php_hydra_globals), NULL, NULL);
	/* hydra_register_on_close( NULL); */
#endif

	if ((ini = getenv("PHP_INI_PATH"))) {
		hydra_sapi_module.php_ini_path_override = ini;
	}

	sapi_startup(&hydra_sapi_module);
	hydra_sapi_module.startup(&hydra_sapi_module);

	TSRMLS_FETCH();
	
}

void _php_hic_shutdown(void)
{
	TSRMLS_FETCH();

	if (SG(server_context) != NULL) {
		hydra_sapi_module.shutdown(&hydra_sapi_module);
		sapi_shutdown();
#ifdef ZTS
		tsrm_shutdown();
#endif
	}
}
